#!/usr/bin/env python3

import argparse
import asyncio
import logging
import sys
from concurrent.futures import ThreadPoolExecutor
from getpass import getuser
from pathlib import Path
from typing import Awaitable, Dict, List
from json2netns.config import Config
from json2netns.netns import Namespace, MacVlan, setup_all_veths, setup_global_oob

GLOBAL_OOB_INTERFACE = "oob0"
LOG = logging.getLogger(__name__)
VALID_ACTIONS = {"create", "delete", "check"}
VALID_SORTED_ACTIONS = sorted(VALID_ACTIONS)


def amiroot() -> bool:
    return getuser() == "root"


async def async_main(args: argparse.Namespace) -> int:
    config = Config(Path(args.config))
    topology_config = config.load()
    executor = ThreadPoolExecutor(max_workers=args.workers)

    namespaces: Dict[str, Namespace] = {}
    for ns_name, ns_config in topology_config["namespaces"].items():
        namespaces[ns_name] = Namespace(ns_name, ns_config, topology_config)

    if not amiroot():
        LOG.error("Please `sudo` / become root to run netns commands")
        return 69

    lower_action = args.action.lower()
    if lower_action == "check":
        ns_count = 0
        for ns_name, ns in namespaces.items():
            if ns_count != 0:
                print("")
            print(f"# Checking {ns_name}")
            ns.check()
            ns_count += 1

        LOG.debug(f"Ran check commands for {ns_count} NSs")
        return 0

    # Perform non co-ro fun
    if lower_action == "create":
        # veth pairs need to be setup then moved to namespaces
        setup_all_veths(namespaces)
        # Add global oob if wanted
        setup_global_oob(GLOBAL_OOB_INTERFACE, namespaces, topology_config)
    elif lower_action == "delete":
        # Check if we have an oob device and clean it up
        oob_int = MacVlan(GLOBAL_OOB_INTERFACE, "deleting_only")
        oob_int.delete()

    # Create NS Coros and run in parallel
    loop = asyncio.get_running_loop()
    namespace_coros: List[Awaitable] = []
    for _ns_name, ns in namespaces.items():
        if lower_action == "create":
            namespace_coros.append(loop.run_in_executor(executor, ns.setup))
        elif lower_action == "delete":
            namespace_coros.append(loop.run_in_executor(executor, ns.delete))

    if not namespace_coros:
        LOG.error(f"Nothing to do. Is {lower_action} a valid action?")
        return 10

    try:
        await asyncio.gather(*namespace_coros)
    except Exception as e:
        LOG.error(f"{sys.argv[0]} Failed: {e}")
        return -1
    return 0


def validate_args(args: argparse.Namespace) -> int:
    """Look at args and make sure that are valid"""
    config_path = Path(args.config)
    if not config_path.exists():
        LOG.error("We need a JSON topology config to do anything")
        return 1

    if args.action.lower() not in VALID_ACTIONS:
        LOG.error(
            f"{args.action} is not a valid action choice! Valid choices: "
            + f"'{' '.join(VALID_SORTED_ACTIONS)}'"
        )
        return 2
    return 0


def main() -> int:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Verbose debug output"
    )
    parser.add_argument(
        "--validate",
        action="store_true",
        help="Validate JSON config (not yet implemented)",
    )
    parser.add_argument(
        "--workers",
        type=int,
        default=1,
        help="Number of threads to for per netns operations",
    )
    parser.add_argument("config", help="Path to JSON topology config")
    parser.add_argument(
        "action", help=f"Action to perform: {'|'.join(VALID_SORTED_ACTIONS)}"
    )
    args = parser.parse_args()

    log_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(
        format="[%(asctime)s] %(levelname)s: %(message)s (%(filename)s:%(lineno)d)",
        level=log_level,
    )
    LOG.debug(f"Starting {sys.argv[0]} ...")
    error_value = validate_args(args)
    if error_value:
        return error_value
    return int(asyncio.run(async_main(args)))


if __name__ == "__main__":  # pragma: nocover
    sys.exit(main())



import logging
from json import load
from pathlib import Path
from typing import Dict


LOG = logging.getLogger(__name__)


class Config:
    """Handle the JSON config file"""

    def __init__(self, path: Path) -> None:
        self.path = path

    def load(self) -> Dict:
        """Load JSON config to use with creating Namespace objects"""
        with self.path.open("rb") as cfp:
            return dict(load(cfp))

    # TODO: Validate config
    def validate(self, config: Dict) -> None:
        """High level config validator"""
        pass



import logging
from ipaddress import IPv4Interface, IPv6Interface, ip_interface, ip_network
from pathlib import Path
from subprocess import CompletedProcess, DEVNULL, PIPE, run
from typing import Dict, List, Optional, Sequence, Union


DEFAULT_IP = "/usr/sbin/ip"
IPInterface = Union[IPv4Interface, IPv6Interface]
LOG = logging.getLogger(__name__)


class Namespace:
    IP = DEFAULT_IP
    check_commands = {
        "## Addresses": (IP, "addr", "show"),
        "## Routes (v4)": (IP, "route", "show"),
        "## Routes (v6)": (IP, "-6", "route", "show"),
    }

    def __init__(self, name: str, ns_config: Dict, config: Dict) -> None:
        self.name = name
        self.ns_path = Path(f"/run/netns/{self.name}")
        self.config = config
        if ns_config["id"] < 1:
            # The global oob and other resources will always be 0
            raise ValueError("A namespace ID must be > 0")
        self.id = ns_config["id"]
        self.interfaces = ns_config["interfaces"]
        self.routes = ns_config["routes"]

        self.oob = ns_config["oob"]
        self.oob_prefixes = None
        if self.oob:
            self.oob_prefixes = [
                ip_network(prefix) for prefix in config["oob"]["prefixes"]
            ]

    def _create_or_delete(
        self, delete: bool, check: bool = True
    ) -> Optional[CompletedProcess]:
        """Create or delete the entire netns"""
        if not delete and self.ns_path.exists():
            LOG.info(f"{self.name} namespace already exists ...")
            return None
        elif delete and not self.ns_path.exists():
            LOG.info(f"{self.name} namespace does not exist ...")
            return None

        op = "delete" if delete else "add"
        d = "d" if delete else "ed"
        cp = run((f"{self.IP}", "netns", op, self.name), check=check)
        LOG.info(f"Successfully {op}{d} {self.name} namespace")
        return cp

    # TODO: Use Interface Objects here
    def _link_add(
        self,
        link_name: str,
        int_type: str,
        *,
        peer_name: str = "",
        physical_int: str = "",
    ) -> None:
        cmd = [self.IP, "link", "add", link_name]
        if int_type == "macvlan":
            cmd.extend(["link", physical_int, "type", "macvlan", "mode", "bridge"])
        elif int_type == "veth":
            return self._link_set_ns(link_name)
        else:
            raise ValueError(f"{int_type} is not supported yet ... PR time?")

        run(cmd, check=True)
        LOG.info(f"Successfully created {link_name}")
        self._link_set_ns(link_name)

    def _link_set_ns(self, link_name: str) -> None:
        link_set_cmd = [self.IP, "link", "set", link_name, "netns", self.name]
        run(link_set_cmd, check=True)
        LOG.info(f"Moved {link_name} to {self.name} namespace")

    def _prefix_add(
        self,
        interface: str,
        prefixes: Sequence[Union[IPInterface, str]],
    ) -> None:
        """Add prefixes onto a namespace interfance"""
        for prefix in prefixes:
            cmd = [self.IP, "addr", "add", str(prefix), "dev", interface]
            self.exec_in_ns(cmd)
            LOG.info(f"Added {str(prefix)} to {interface} in {self.name} namespace")

    def check(self) -> None:
        for header, cmd in self.check_commands.items():
            print(header)
            self.exec_in_ns(cmd, check=False)

    def create(self, delete: bool = False) -> None:
        self._create_or_delete(delete=False)

    def delete(self) -> None:
        self._create_or_delete(delete=True)

    def create_devices_and_address(self) -> None:
        """Create virtual network device and assign to the netns"""
        for int_name, int_config in self.interfaces.items():
            if int_config["type"] == "loopback":
                continue

            if self.interface_exists(int_name) == 0:
                LOG.info("Skipping interface creation as {int_name} already exits")
                continue

            peer_name = int_config["peer_name"] if "peer_name" in int_config else ""
            physical_int = (
                self.config["physical_int"] if "physical_int" in self.config else ""
            )
            self._link_add(
                int_name,
                int_config["type"],
                peer_name=peer_name,
                physical_int=physical_int,
            )
            self._prefix_add(int_name, int_config["prefixes"])
            self.up_interface(int_name)

    def create_oob(self) -> None:
        """If configured, add a OOB device to connect to global namespace
        + bridge with a physical interface"""
        if not self.oob:
            LOG.debug(f"No oob configred for {self.name}")
            return

        oob_int_name = f"oob{self.id}"
        physical_int = (
            self.config["physical_int"] if "physical_int" in self.config else ""
        )
        if not physical_int:
            raise ValueError(
                f"No Physical int to bridge macvlan OOB interface with for {self.name}"
            )
        self._link_add(oob_int_name, "macvlan", physical_int=physical_int)
        oob_prefixes = self.oob_addrs()
        self._prefix_add(oob_int_name, oob_prefixes)

    def exec_in_ns(
        self, cmd: Sequence[str], check: bool = True, output: bool = True
    ) -> CompletedProcess:
        """Run command from inside the netns"""
        output_fd = None if output else DEVNULL
        ns_cmd = [self.IP, "netns", "exec", self.name]
        ns_cmd.extend(cmd)
        LOG.debug(f"Running '{' '.join(ns_cmd)}' in {self.name} namespace")
        cp = run(ns_cmd, check=check, stdout=output_fd, stderr=output_fd)
        LOG.debug(
            f"Finished running '{' '.join(ns_cmd)}' {self.name} namespace "
            + f"(returned {cp.returncode})"
        )
        return cp

    def interface_exists(self, interface: str, outside_ns: bool = False) -> int:
        """Check if a device exists"""
        cmd = [self.IP, "link", "show", "dev", interface]
        if outside_ns:
            return run(cmd, stdout=DEVNULL, stderr=DEVNULL).returncode
        else:
            return self.exec_in_ns(cmd, check=False, output=False).returncode

    def oob_addrs(self) -> List[IPInterface]:
        if not self.oob:
            LOG.error(f"No oob asked to be set on {self.name} namespace")
            return []
        if not self.oob_prefixes:
            LOG.error(f"No oob prefiex to apply to {self.name}")
            return []

        interfaces = []
        for oob_net in self.oob_prefixes:
            interfaces.append(
                ip_interface(
                    f"{str(oob_net.network_address + self.id)}/{int(oob_net.prefixlen)}"
                )
            )

        return interfaces

    # TODO: Add support
    def route_add(self) -> None:
        """Add any needed static routes in netns"""
        pass

    def up_interface(self, interface: str) -> None:
        up_cmd = [self.IP, "link", "set", "up", "dev", interface]
        self.exec_in_ns(up_cmd, output=False)
        LOG.debug(f"Set '{interface}' interface up")

    def setup_loopback(self, loopback_int_name: str = "lo") -> None:
        """Set loopback up (always) + add prefixes if desired"""
        self.up_interface(loopback_int_name)
        if (
            loopback_int_name not in self.interfaces
            and "prefixes" not in self.interfaces[loopback_int_name]
        ):
            return
        prefixes = []
        for prefix in self.interfaces[loopback_int_name]["prefixes"]:
            prefixes.append(ip_interface(prefix))
        if prefixes:
            self._prefix_add(loopback_int_name, prefixes)

    def setup(self) -> None:
        """Cordination function to setup all the namespace elements
        - Designed to be idempotent - i.e. if it exists, move on"""
        # Create netns
        self.create()
        # Always bring loopback up
        self.setup_loopback()
        # Create links/interfaces + address them
        self.create_devices_and_address()
        # Create oob if selected
        self.create_oob()
        # Add any static routes
        self.route_add()


class Interface:
    IP = DEFAULT_IP
    name = "Interface"
    type = "Interface"

    def add_prefixes(self, prefixes: Sequence[IPInterface]) -> None:
        for prefix in prefixes:
            cmd = [self.IP, "addr", "add", str(prefix), "dev", self.name]
            run(cmd, check=True, stdout=PIPE, stderr=PIPE)

    def delete(self) -> Optional[CompletedProcess]:
        if not self.exists():
            LOG.debug(
                f"Not deleting {self.name} {self.type} interface as it does not exist ..."
            )
            return None

        cmd = [self.IP, "link", "del", self.name]
        LOG.info(f"Deleting {self.type} interface {self.name}")
        return run(cmd, check=True, stdout=PIPE, stderr=PIPE)

    def exists(self) -> bool:
        """Check if a interface device exists"""
        cmd = [self.IP, "link", "show", "dev", self.name]
        return run(cmd, stdout=DEVNULL, stderr=DEVNULL).returncode == 0


class MacVlan(Interface):
    """Class to create macvlan interfaces Prefix assignment + interface creation"""

    def __init__(self, name: str, physical_int: str, *, mode: str = "bridge") -> None:
        self.name = name
        self.physical_interface = physical_int
        self.type = "macvlan"
        self.mode = mode

    def create(self) -> CompletedProcess:
        cmd = [
            self.IP,
            "link",
            "add",
            self.name,
            "link",
            self.physical_interface,
            "type",
            self.type,
            "mode",
            self.mode,
        ]
        LOG.info(
            f"Created {self.type} {self.name} bridged to {self.physical_interface}"
        )
        return run(cmd, check=True)


class Veth(Interface):
    """Veths pairs need to be made in default namespace then moved to NetNS"""

    def __init__(self, name: str, peer: str) -> None:
        self.name = name
        self.peer = peer
        self.type = "veth"

    def create(self) -> CompletedProcess:
        cmd = [self.IP, "link", "add", self.name, "type", self.type, "peer", self.peer]
        LOG.info(f"Created veth {self.name} with peer {self.peer}")
        return run(cmd, check=True)


def setup_all_veths(namespaces: Dict[str, Namespace]) -> None:
    """Setup all veths in a namespace then move to netns where needed"""
    for _ns_name, ns in namespaces.items():
        LOG.debug(f"Checking veths for {ns.name} namespace")
        for int_name, int_config in ns.interfaces.items():
            if int_config["type"] != "veth":
                continue

            if (
                ns.interface_exists(
                    int_name,
                    outside_ns=True,
                )
                == 0
                or (ns.ns_path.exists() and ns.interface_exists(int_name))
            ):
                LOG.debug(f"Not setting up {int_name} as it exists")
                continue

            veth = Veth(int_name, int_config["peer_name"])
            veth.create()
            LOG.info(f"Created {int_name} veth")


def setup_global_oob(
    interface_name: str, namespaces: Dict[str, Namespace], config: Dict
) -> None:
    """Add Global OOB interface if any netns has oob set to true"""
    oob_wanted = False
    for _, ns in namespaces.items():
        if ns.oob:
            oob_wanted = True
    if not oob_wanted:
        LOG.debug(
            "Not setting up a global OOB device. No namespace has an oob interface"
        )
        return

    oob_int = MacVlan(interface_name, config["physical_int"])
    oob_int.create()



#!/usr/bin/env python3

import argparse
import asyncio
import unittest
from unittest.mock import patch
from pathlib import Path

import json2netns.main
from json2netns.config import Config
from json2netns.tests.netns import InterfaceTests, NetNSTests  # noqa: F401


BASE_PATH = Path(__file__).parent.parent.resolve()
SAMPLE_CONF = BASE_PATH / "sample.json"


class MainTests(unittest.TestCase):
    def test_amiroot(self) -> None:
        self.assertFalse(json2netns.main.amiroot())

    def test_async_main_check(self) -> None:
        ns = argparse.Namespace(
            action="check", config=str(SAMPLE_CONF), debug=True, workers=1
        )
        with patch("json2netns.netns.Namespace.check") as mock_check, patch(
            "json2netns.main.print"
        ) as mock_print, patch("json2netns.main.amiroot", returm_value=True):
            self.assertEqual(0, asyncio.run(json2netns.main.async_main(ns)))
            self.assertEqual(2, mock_check.call_count)
            self.assertEqual(3, mock_print.call_count)

    def test_main(self) -> None:
        ns = argparse.Namespace(
            action="check", config=str(SAMPLE_CONF), debug=True, workers=1
        )
        with patch(
            "argparse.ArgumentParser.parse_args", return_value=ns
        ) as mock_pa, patch(
            "json2netns.main.async_main", return_value=0
        ) as mock_async_main:
            self.assertEqual(0, json2netns.main.main())
            self.assertEqual(1, mock_async_main.call_count)
            self.assertEqual(1, mock_pa.call_count)

    def test_bad_args(self) -> None:
        ns = argparse.Namespace(
            action="check", config=str(BASE_PATH / "not_there"), debug=True, workers=1
        )
        self.assertEqual(1, json2netns.main.validate_args(ns))

        ns.action = "cooper_fuck_yes"
        ns.config = str(SAMPLE_CONF)
        self.assertEqual(2, json2netns.main.validate_args(ns))


class ConfigTests(unittest.TestCase):
    def setUp(self) -> None:
        self.config = Config(SAMPLE_CONF)

    def test_load_config(self) -> None:
        self.assertTrue(isinstance(self.config.load(), dict))


if __name__ == "__main__":
    unittest.main()



from setuptools import setup

from pathlib import Path


CURRENT_DIR = Path(__file__).parent


# Specific Python Test Runner (ptr) params for Unit Testing Enforcement
ptr_params = {
    "entry_point_module": "src/json2netns/main",
    "test_suite": "json2netns.tests.base",
    "test_suite_timeout": 120,
    "required_coverage": {
        "json2netns/config.py": 90,
        "json2netns/main.py": 70,
        "json2netns/netns.py": 70,
    },
    "run_black": True,
    "run_mypy": True,
    "run_flake8": True,
}


def get_long_description() -> str:
    return (CURRENT_DIR / "README.md").read_text(encoding="utf8")


setup(
    name="json2netns",
    version="2021.5.17",
    description="JSON parsing Linux Network Namespace (netns) topology builder.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    keywords="json linux netns network namespace",
    author="Cooper Ry Lees",
    author_email="me@cooperlees.com",
    url="https://github.com/cooperlees/json2netns",
    license="BSD",
    packages=["json2netns", "json2netns.tests"],
    package_dir={"": "src"},
    package_data={
        "json2netns": ["sample.json"],
    },
    python_requires=">=3.8.0",
    install_requires=[],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    entry_points={
        "console_scripts": [
            "json2netns=json2netns.main:main",
        ]
    },
)



#!/usr/bin/env python3

import unittest
from copy import deepcopy
from ipaddress import IPv4Interface, IPv6Interface
from pathlib import Path
from unittest.mock import patch

from json2netns.config import Config
from json2netns.netns import Interface, MacVlan, Namespace, Veth, setup_all_veths

BASE_PATH = Path(__file__).parent.parent.resolve()
BASE_MODULE = "json2netns.netns"
SAMPLE_JSON_CONF_PATH = BASE_PATH / "sample.json"


class InterfaceTests(unittest.TestCase):
    def setUp(self) -> None:
        self.interface = Interface()
        self.mac_vlan = MacVlan("macvlan0", "eth0")
        self.veth = Veth("veth0", "veth69")

    def test_delete(self) -> None:
        with patch(f"{BASE_MODULE}.run") as mock_run:
            self.assertIsNone(self.interface.delete())
            self.assertEqual(1, mock_run.call_count)


class NetNSTests(unittest.TestCase):
    def setUp(self) -> None:
        self.config = Config(SAMPLE_JSON_CONF_PATH).load()
        for ns_name, ns_conf in self.config["namespaces"].items():
            self.test_ns = Namespace(ns_name, ns_conf, self.config)
            break

    def test_bad_id(self) -> None:
        bad_config = deepcopy(self.config)
        bad_config["namespaces"]["left"]["id"] = 0
        with self.assertRaises(ValueError):
            Namespace("bad_id", bad_config["namespaces"]["left"], bad_config)

    def test_check(self) -> None:
        with patch(f"{BASE_MODULE}.run") as mock_run, patch(
            f"{BASE_MODULE}.print"
        ) as mock_print:
            self.test_ns.check()
            # Mocked objects should be both called for each check command
            expected_calls = len(self.test_ns.check_commands)
            self.assertEqual(expected_calls, mock_run.call_count)
            self.assertEqual(expected_calls, mock_print.call_count)

    def test_create(self) -> None:
        with patch(f"{BASE_MODULE}.run") as mock_run:
            self.test_ns.create()
            self.assertEqual(1, mock_run.call_count)

    def test_create_devices_and_address(self) -> None:
        with patch(f"{BASE_MODULE}.run") as mock_run, patch(
            f"{BASE_MODULE}.Namespace.exec_in_ns"
        ) as mock_exec_ns:
            self.test_ns.create_devices_and_address()
            # Only 1 non loopback veth interface - 1 call to just setns
            self.assertEqual(1, mock_run.call_count)
            # 4 Calls:
            # - Check if exists
            # - 2 for adding the v6 and v6 prefixes
            # - Setting interface 'up'
            self.assertEqual(4, mock_exec_ns.call_count)

    def test_delete(self) -> None:
        with patch(f"{BASE_MODULE}.LOG.info") as mock_log, patch(
            f"{BASE_MODULE}.run"
        ) as mock_run:
            self.test_ns.delete()
            self.assertEqual(1, mock_log.call_count)
            # We should not delete as it should not exist
            self.assertEqual(0, mock_run.call_count)

    def test_valid_class(self) -> None:
        self.assertTrue("left", self.test_ns.name)

    def test_oob_addrs(self) -> None:
        expected = [IPv6Interface("fddd::1/64"), IPv4Interface("10.255.255.1/24")]
        self.assertEqual(expected, self.test_ns.oob_addrs())

    def test_setup_all_veths(self) -> None:
        # This test can not be ran in an env where a left0 interface can exist
        ns_dict = {"left": self.test_ns}
        with patch(f"{BASE_MODULE}.run") as mock_run:
            setup_all_veths(ns_dict)
            # Once run call for check interfaces exists and one for the veth add
            self.assertEqual(2, mock_run.call_count)
